// MQL4&5-code

// #define HISTORYTICKS_ON //   HistoryTicks

#property strict

#ifndef _CS
  #define _CS(A) (A)
#endif // #ifndef _CS

#ifdef HISTORYTICKS_ON

#include "ArrayResize.mqh"
#include "Data_String.mqh"
#include "Array.mqh"
#include "ResourceData.mqh"

MqlTick HistoryTicks[];
MqlTick LastTicks[];

#define NAME "HistoryTicks_" + _Symbol

#define RESERVE_HISTORY 100000
#define RESERVE_LAST 100

#define ALERT(A) ::Alert(::MQLInfoString(MQL_PROGRAM_NAME) + ": " + __FUNCSIG__ + ", Line = " + (string)__LINE__ + ": " + #A + " - Error!")

class HISTORYTICKS
{
private:
  static const RESOURCEDATA<MqlTick> Resource;
//  static const bool LoadTicks;
  static MqlTick PrevTick;
  static int PrevAmountTicks;

#define FILENAME (__FILE__ + ".tpl")
#define PATH "\\Files\\"

  // https://www.mql5.com/ru/docs/files/fileload
  static long FileLoad( const string FileName, uchar &Buffer[], const int CommonFlag = 0 )
  {
    long Res = -1;

    const int handle = ::FileOpen(FileName, FILE_READ | FILE_BIN | CommonFlag);

    if (handle != INVALID_HANDLE)
    {
      if (!(Res = ::FileReadArray(handle, Buffer)))
      {
        ALERT(::FileReadArray(handle, Buffer));

        Res = -1;
      }

      ::FileClose(handle);
    }

    return(Res);
  }

  static string TemplateToString( const long Chart_ID = 0 )
  {
    uchar Data[];

    return((::ChartSaveTemplate((ulong)Chart_ID, PATH + FILENAME) && (HISTORYTICKS::FileLoad(FILENAME, Data) > 0)) ? ::CharArrayToString(Data) : NULL);
  }

  static bool IsRun( void )
  {
    bool Res = false;

    for (long chartID = ::ChartFirst(); _CS((chartID != -1) && !Res); chartID = ::ChartNext(chartID))
      Res = HISTORYTICKS::IsRun(chartID);

    ::FileDelete(FILENAME);

    if (!Res)
      ::Alert("Run the " + NAME + "-indicator!!!");

    return(Res);
  }

#undef PATH
#undef FILENAME

  static int GetHistoryTicks( void )
  {
    // https://www.mql5.com/ru/forum/1111/page2336#comment_9641032
    const string Name = NAME;
    const string TransferName = Name + "_Transfer";

    if (!::GlobalVariableSet(TransferName, (int)::GlobalVariableGet(TransferName) + 1))
      ALERT(::GlobalVariableSet(TransferName, (int)::GlobalVariableGet(TransferName) + 1));

    ::Comment("Waiting HistoryTicks..." +
            (::GlobalVariableCheck(Name) ? NULL : "\nRun the " + Name + "-indicator!!!"));

    //  ,     
    ArrayFree(::HistoryTicks);
    ArrayFree(::LastTicks);
    HISTORYTICKS::PrevTick.time = INT_MAX; // . HISTORYTICKS::OnChartEvent

    MqlTick PrevTick2 = {0};
    ::SymbolInfoTick(_Symbol, PrevTick2);

    MqlTick NewTick;
    ulong StartTime = ::GetMicrosecondCount();

    while (!::IsStopped() && !HISTORYTICKS::Resource.Get(::HistoryTicks))
      if (PrevTick2.time && ::SymbolInfoTick(_Symbol, NewTick) && _R(NewTick) != PrevTick2) //      ,  .
      {
        PrevTick2.time = 0;

        StartTime = ::GetMicrosecondCount();
      }

    StartTime = ::GetMicrosecondCount() - StartTime;

    const int Value = (int)::GlobalVariableGet(TransferName);

    if (Value == 1)
    {
      if (!::GlobalVariableDel(TransferName))
        ALERT(::GlobalVariableDel(TransferName));
    }
    else if (!::GlobalVariableSet(TransferName, Value - 1))
      ALERT(::GlobalVariableSet(TransferName, Value - 1));

    ::Comment("");

    const int AmountTicks = ::ArraySize(::HistoryTicks);

    if (AmountTicks)
      ::Print("Received  " + (string)AmountTicks + " ticks (" + ::DoubleToString((long)AmountTicks * sizeof(MqlTick) / (double)(1 << 20), 2) + " MB): " +
              (string)::HistoryTicks[0].time + " - " + (string)::HistoryTicks[AmountTicks - 1].time +
              ", TransferTime = " + (string)(StartTime / 1000) + "ms." +
              (StartTime ? " (" + (string)((long)AmountTicks * 1000000 / StartTime) + " ticks/sec - " +
              ::DoubleToString(((long)AmountTicks * 1000000 / StartTime) * sizeof(MqlTick) / (double)(1 << 20), 2) + " MB/sec)": NULL));
    else
      ::Alert("Received 0 ticks! TransferTime = " + (string)(StartTime / 1000) + " ms.");

    return(AmountTicks);
  }

public:
  static bool IsRun( const long Chart_ID )
  {
    // MQL4 - StringFind  -1  .
    return(::StringFind(HISTORYTICKS::TemplateToString(Chart_ID), "<indicator>\r\nname=Custom Indicator\r\n<expert>\r\nname=HistoryTicks") != -1);
  }

  static bool IsChange( const bool Full = true )
  {
    static int PrevValue = 0;
    static MqlTick PrevTick2 = {0};

    const int Value = Full ? HISTORYTICKS::GetAmountTicks() : HISTORYTICKS::GetPrevAmountTicks();
    const bool Res1 = Value ? Value != PrevValue : false;

    MqlTick Tick;

  #ifdef __VIRTUAL__
    const bool Res2 = _V(0, ::SymbolInfoTick(_Symbol, Tick) && (_R(PrevTick2) != Tick));
  #else // __VIRTUAL__
    const bool Res2 = ::SymbolInfoTick(_Symbol, Tick) && (_R(PrevTick2) != Tick);
  #endif // __VIRTUAL__

    const bool Res = Res1 || Res2;

    if (Res)
    {
      PrevValue = Value;
      PrevTick2 = Tick;
    }

    return(Res);
  }

  // https://www.mql5.com/ru/forum/1111/page2961#comment_20811791
  static int GetAmountTicks( void )
  {
    static const string Name = NAME;

//    return((int)::GlobalVariableGet(Name)); //   .

  #ifdef __MQL5__
    static const RESOURCE Counter("\\Indicators\\HistoryTicks.ex5::" + Name + "_Counter");
  #else // __MQL5__
    static const RESOURCE Counter("\\Indicators\\HistoryTicks.ex4::" + Name + "_Counter");
  //  static const RESOURCE Counter("\\Indicators\\HistoryTicks_Beta.ex4::" + Name + "_Counter");
  #endif // __MQL5__

    uint Data[];
    const int Size = Counter.Get(Data);

    if (Size <= 0)
      ALERT(Counter.Get(Data));

    return(HISTORYTICKS::PrevAmountTicks = ((Size > 0) ? (int)Data[0] : 0));
  }

  static int GetPrevAmountTicks( void )
  {
    return(HISTORYTICKS::PrevAmountTicks);
  }

/*
  static bool InitTick( MqlTick &Tick, const datetime time )
  {
    Tick.time = time;

    return(true);
  }
*/
  static void OnChartEvent( const int id, const long &lparam, const double &dparam, const string &sparam )
  {
    static const string Name = NAME;
    static MqlTick Tick = {0};

    if (id == CHARTEVENT_CUSTOM)
    {
      Tick = DATA_STRING::FromString<MqlTick>(sparam);

      if ((lparam > ::ArraySize(::HistoryTicks)) ||  //   ,       .
          (Tick.time > HISTORYTICKS::PrevTick.time)) //    ,      .
      {
        if (::ArraySize(::HistoryTicks)) //     ArrayFree(HistoryTicks),     .
          ARRAY::AddElement(::HistoryTicks, Tick, RESERVE_HISTORY);

        ARRAY::AddElement(::LastTicks, Tick, RESERVE_LAST);

        HISTORYTICKS::PrevTick = Tick;
      }

      if ((lparam >= HISTORYTICKS::GetAmountTicks()) && ::ArraySize(::LastTicks))
      {
        ::OnTick2();

        ::ArrayFree(::LastTicks);
      }
    }
  }

  static void OnTick( void )
  {
    static const string Name = NAME;

    if (!::GlobalVariableCheck(Name))
      ::Print("Run the " + Name + "-indicator!!!");

    return;
  }

  static int OnInit( void )
  {
    const bool Res = HISTORYTICKS::IsRun();

    if (Res)
      HISTORYTICKS::GetHistoryTicks();

    return(Res ? OnInit2() : INIT_FAILED);
  }
};

#ifdef __MQL5__
  static const RESOURCEDATA<MqlTick> HISTORYTICKS::Resource("\\Indicators\\HistoryTicks.ex5::" + NAME);
#else // __MQL5__
  static const RESOURCEDATA<MqlTick> HISTORYTICKS::Resource("\\indicators\\HistoryTicks.ex4::" + NAME);
//  static const RESOURCEDATA<MqlTick> HISTORYTICKS::Resource("\\indicators\\HistoryTicks_Beta.ex4::" + NAME);
#endif // __MQL5__

static MqlTick HISTORYTICKS::PrevTick = {0};
// static const bool HISTORYTICKS::LoadTicks = HISTORYTICKS::GetHistoryTicks(); // -    Resource    GetHistoryTicks,  .
static int HISTORYTICKS::PrevAmountTicks = 0;

#undef ALERT

#undef RESERVE_LAST
#undef RESERVE_HISTORY

#undef NAME

void OnChartEvent( const int id, const long &lparam, const double &dparam, const string &sparam )
{
  HISTORYTICKS::OnChartEvent(id, lparam, dparam, sparam);

  return;
}

void OnTick( void )
{
  HISTORYTICKS::OnTick();
}

int OnInit( void )
{
  return(HISTORYTICKS::OnInit());
}

#define OnTick OnTick2
#define OnInit OnInit2

#endif // HISTORYTICKS_ON